/*!
  \file Application.cpp

  \brief Application object implementation


  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   17.03.2006 11:32:26 MSD
*/

#include "h/Application.h"

#include <boost/tokenizer.hpp>

#include <set>
#include <algorithm>

using namespace Ice;

/*!
  \typedef Separator_t

  \brief Defines tokenizer's char separator
*/
typedef boost::char_separator<char> Separator_t;

/*!
  \typedef Tokenizer_t

  \brief Defines tokenizer with Separator_t as separator type
*/
typedef boost::tokenizer<Separator_t> Tokenizer_t;

/*!
  \brief Calculates number of tokens in expression

  \param expression - String of tokens
  \param sep - separator
*/
unsigned TokSize(const std::string expression, Separator_t sep)
{
  unsigned size = 0;
  Tokenizer_t tok(expression,sep);
  for( Tokenizer_t::iterator i=tok.begin(); i!=tok.end();++i) size++;
  return size;
}

void Ice::DummyAppStreamCreatedHandler(AppStream&) 
{
  // Do nothing	
}

void Ice::DummySimpleHandler(const std::string& msg)
{
  // Do nothing
}

void Ice::DummyErrorHandler(const int& code, const std::string& hr_code, const std::string& msg)
{
  // Do nothing
}

//************************************ ApplicationAttrsDesc_t ***********************************************//
ApplicationAttrsDesc_t::AttrsDescription()
{
  Add(apConnectable,  "CONNECTABLE");
  Add(apConnecting,   "CONNECTING");
  Add(apStreams,      "STREAMS");
  Add(apSending,      "SENDING");
  Add(apReceived,     "RECEIVED");
}	
//************************************ Application **********************************************************//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
Application::Application(SkypeConnection& _sc, const std::string& _app_name, const LogFunc_t& _log_func, const  SimpleHandler_t& _sh, const ErrorHandler_t& _erh)
	   : Parent_t(_sc, *this, "APPLICATION", _app_name, _log_func), 
	     sc(_sc), name(_app_name), isCreated(false), isConnected(false),
	     errorHandler(_erh), simpleHandler(_sh)
{ 
  Init(); 
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////

Application::Application(SkypeConnection& _sc, const std::string& _app_name, const ObjErrHandler_t& _eh, const LogFunc_t& _log_func, const SimpleHandler_t& _sh, const ErrorHandler_t& _erh)
	   : Parent_t(_sc, *this, "APPLICATION", _app_name, _eh, _log_func), 
	     sc(_sc), name(_app_name), isCreated(false), isConnected(false), 
	     errorHandler(_erh), simpleHandler(_sh)
{
  Init(); 
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::Init()
{	
  regExps_[genericRE] = std::string("^\\s*APPLICATION\\s+") + name + std::string("\\s+(\\w+)\\s+(.*)\\s*$");
  
  struct Parser_t::SkypeObjRegEx re;
  re.resp = boost::regex(regExps_[genericRE].c_str());
  re.resp_value_pos = 2;
  re.event_attr_pos = 1;
  re.event_value_pos = 2;
  parser.SetRegEx(re);
  regExpsIds_[genericRE] = sc.RegisterCallback(regExps_[genericRE].c_str(), RegExNotify_t(this, &Application::ProcessEvent));	
  
  regExps_[onCreateRE] = "^\\s*CREATE\\s+APPLICATION\\s+" + name + "\\s*$";
  regExpsIds_[onCreateRE] = sc.RegisterCallback(regExps_[onCreateRE].c_str(), RegExNotify_t(this, &Application::onCreate));	
  regExps_[onDeleteRE] = "^\\s*DELETE\\s+APPLICATION\\s+" + name + "\\s*$";
  regExpsIds_[onDeleteRE] = sc.RegisterCallback(regExps_[onDeleteRE].c_str(), RegExNotify_t(this, &Application::onDelete));	
  regExps_[onStreamRE] = "^\\s*APPLICATION\\s+" + name + "\\s+STREAMS\\s+(.*)$";
  regExpsIds_[onStreamRE] = sc.RegisterCallback(regExps_[onStreamRE].c_str(), RegExNotify_t(this, &Application::onStream));	
  regExps_[onReceivedRE] = "^\\s*APPLICATION\\s+" + name + "\\s+RECEIVED\\s+(.+)\\d+$";
  regExpsIds_[onReceivedRE] = sc.RegisterCallback(regExps_[onReceivedRE].c_str(), RegExNotify_t(this, &Application::onReceived));	
  regExps_[onReadRE] = "^\\s*ALTER\\s+APPLICATION\\s+" + name + "\\s+READ\\s([^\\s]+)\\s(.+)$";
  regExpsIds_[onReadRE] = sc.RegisterCallback(regExps_[onReadRE].c_str(), RegExNotify_t(this, &Application::onRead));	
  regExps_[onConnectedRE] = "^\\s*APPLICATION\\s+" + name + "\\s+CONNECTING$";
  regExpsIds_[onConnectedRE] = sc.RegisterCallback(regExps_[onConnectedRE].c_str(), RegExNotify_t(this, &Application::onConnected));
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
Application::~Application()
{
  typedef RegExps_t::iterator IT;
  for(IT i = regExps_.begin(); i!=regExps_.end();i++)
    {
       assert(regExpsIds_.find(i->first) != regExpsIds_.end());
       sc.UnRegisterCallback(i->second, regExpsIds_[i->first]);     
    }  
  for(StreamsMap_t::iterator i = streams_.begin(); i!=streams_.end();i++)
    {
       sc << DisableIds << "ALTER APPLICATION " << name << " DISCONNECT " << i->first << flush << EnableIds;
    }
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::Create()
{
  sc <<  "CREATE APPLICATION " << name << Flush(simpleHandler,errorHandler);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::Delete()
{
  sc <<  "DELETE APPLICATION " << name << Flush(simpleHandler,errorHandler);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::onCreate(const boost::cmatch& _event)
{
  isCreated = true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::onDelete(const boost::cmatch& _event)
{
  isCreated = false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::onConnected(const boost::cmatch& _event)
{
  isConnected = true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::onReceived(const boost::cmatch& _event)
{
  const std::string streams = _event[1];
  Separator_t sep(" ");
  Tokenizer_t tok(streams,sep);
  /// Loop by all streams in event
  for(Tokenizer_t::iterator i=tok.begin(); i!=tok.end();++i)
    {
      std::string stream = std::string((*i).begin(),std::find((*i).begin(), (*i).end(), '='));
      log(LlDbg,"STREAM = '%s'", stream.c_str());	    
      assert(streams_.find(stream) != streams_.end() && "'RECEIVED' was received, but no stream in storage");
      /// Read data from stream
      sc << "ALTER APPLICATION " << name << " READ " << stream << Flush(simpleHandler,errorHandler);
    }
  isStreamEmpty = false;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::onStreamEmpty(const boost::cmatch& _event)
{
  isStreamEmpty = true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::onRead(const boost::cmatch& _event)
{
  const std::string stream = _event[1];
  const std::string data   = _event[2];
  streams_[stream]->append(data);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::onStream(const boost::cmatch& _event)
{
  const std::string event = _event[1].str();	
  /// Determining number of streams in event
  Separator_t sep(" ");
  Tokenizer_t tok(event,sep);
  unsigned size = TokSize(event,sep);
  if ( streams_.size() < size )
   {
      /// *** Creation ***
      /// Loop by all streams in event
      for(Tokenizer_t::iterator i=tok.begin(); i!=tok.end();++i)
       {
         /// If new stream	   
         if (streams_.find(*i) == streams_.end())
           {
             log(LlInfo,"NEW STREAM = '%s'",(*i).c_str());
             streams_[*i] = new AppStream(*this,*i,sc.GetLog());
	     callback_(*streams_[*i]);
	     //break; //!!!
           }
       }      
   }
  else if ( streams_.size() > size )
   {
      /// *** Deletion ***
      std::set<std::string> tmp;
      /// Loop by all streams in event
      for(Tokenizer_t::iterator i=tok.begin(); i!=tok.end();++i) tmp.insert(*i);
      for(StreamsMap_t::iterator i = streams_.begin(); i != streams_.end(); /*EMPTY*/)
      if ( tmp.find(i->first) == tmp.end() ) 
         {
           log(LlDbg,"DELETE STREAM '%s'",i->first.c_str());
	   streams_.erase(i++);      
	   //break;
	 }
      else ++i;
   }
  else assert (0 && "streams_.size() == size");
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::SetCbk(const AppStreamCreated_t& _h)
{
  callback_ = _h;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void Application::Connect(const std::string& _name, const AppStreamCreated_t& _h)
{
  callback_ = _h;
  sc << "ALTER APPLICATION " << name << " CONNECT " << _name << Flush(simpleHandler,errorHandler);
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//************************************ AppFlush *************************************************************//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
AppFlush::AppFlush(SendMode_t _mode)
	: mode(_mode)
{
};
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
const AppFlush::SendMode_t& AppFlush::GetMode() const
{
  return mode;	
};	
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
//************************************ AppStream ************************************************************//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
AppStream::AppStream(Application& _app, const std::string& stream_id, const LogFunc_t& _log_func)
	  : LogAbility(_log_func), sc(_app.GetSkype()), app(_app), name(stream_id), canRead(false) 
{
};	
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
AppStream& AppStream::operator<<(const struct AppFlush& _f)
{
  return flush(_f.GetMode());	
};	
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
AppStream& AppStream::flush(const AppFlush::SendMode_t& _mode)
{
  std::string msg;
  while (!outBuf.empty()) 
    {
      msg+=outBuf.front();
      outBuf.pop();
    }
  log(LlInfo,"To flush '%s'",msg.c_str());
  switch (_mode)
    {
      case AppFlush::Datagram:
        sc << DisableIds << "ALTER APPLICATION " << app.Name() << " DATAGRAM " << name << " " << msg << ::flush << EnableIds;
        break; 	   
      case AppFlush::Stream:
        sc << DisableIds << "ALTER APPLICATION " << app.Name() << " WRITE " << name << " " << msg << ::flush << EnableIds;
        break; 	   
      default: 
	assert(0 && "Invalid sending mode");	      
    };
  return *this;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool AppStream::CanRead() const
{
  return canRead;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
AppStream& AppStream::operator<<(const char* _buf)
{
  outBuf.push(_buf);
  return *this;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
AppStream& AppStream::operator<<(const std::string& _buf)
{
  outBuf.push(_buf);
  return *this;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void AppStream::append(const std::string& _buf)
{
  inBuf.push(_buf);
  canRead = true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
void AppStream::append(const char* _buf)
{
  inBuf.push(_buf);
  canRead = true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
AppStream& AppStream::operator>>(std::string& _buf)
{
  if (!inBuf.empty()) 
    {
      _buf.clear();
    }
  if(!inBuf.empty())
    {
      _buf += inBuf.front();
      inBuf.pop();
    }
  if(inBuf.empty()) canRead = false;
  
  return *this;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
