/*!
  \file Application.h

  \brief  Application object 
  

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   17.03.2006 10:30:18 MSD
*/

#ifndef APPLICATION_H
#define APPLICATION_H

#include "SkypeObj.h"

#include <queue>

namespace Ice {

/// Application attributes 
typedef enum {
	apConnectable,
	apConnecting,
	apStreams,
	apSending,
	apReceived
} ApplicationAttrs_t;

/// Application attributes description
typedef AttrsDescription<ApplicationAttrs_t> ApplicationAttrsDesc_t;

/// Describes atribute of Skype object "Application"
template <>
class AttrsDescription<ApplicationAttrs_t> : public AttrsDescriptionBase<ApplicationAttrs_t> {
public:
	/// Fills description
	AttrsDescription();
};

// Forward declaration of the application stream class
class AppStream;

/*!
  \brief Functor of this type is called upon new Application Stream creation.
  
  \param New application stream 
*/
typedef Loki::Functor<void, LOKI_TYPELIST_1(AppStream&)> AppStreamCreated_t;

/*!
  \brief This is default for AppStreamCreated_t functor.
  
  \param app_stream - New application stream 
*/
void DummyAppStreamCreatedHandler(AppStream& app_stream);

/*!
  \brief This is default for SimpleHandler_t functor.
  
  \param msg - Message
*/
void DummySimpleHandler(const std::string& msg);

/*!
  \brief This is default for ErrorHandler_t functor.
  
  \param code - Integer skype error code
  \param hr_code - Human-readable error message generated in accordance with Skype error code
  \param msg - Full Skype error message received from Skype
*/
void DummyErrorHandler(const int& code, const std::string& hr_code, const std::string& msg);

/*!
  \class Application

  \brief This class represents application.
*/
class Application : public SkypeObject<Application, ApplicationAttrs_t, SkypeIdObjCmdFormatter,SkypeObjRegExParser,std::string> 
{
public:
	/*!
  	  \brief  Constructs Application object with DEFAULT error handler.  
  		  
	  	  Default error handler doesn't handle any error at all.
		  
  	  \param _sc 		Reference to the Skype Connection
  	  \param _app_name      Application name
	  \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
	  \param _sh            Simple handler
	  \param _erh           Error handler
	*/
	Application(SkypeConnection&       _sc, 
                    const std::string&     _app_name, 
		    const LogFunc_t&       _log_func = DummyLogger, 
		    const SimpleHandler_t& _sh       = SimpleHandler_t(DummySimpleHandler), 
		    const ErrorHandler_t&  _erh      = ErrorHandler_t(DummyErrorHandler)
		   );
	/*!
	  \brief  Constructs Application object with NON-DEFAULT (user-supplied) error handler.  

	  \param _sc		Reference to Skype Connection
  	  \param _app_name      Application name
	  \param _eh		Error hanlder.  This functor will be called in case of errors.
	  \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
	  \param _sh            Simple handler
	  \param _erh           Error handler
	*/
	Application(SkypeConnection&       _sc, 
                    const std::string&     _app_name, 
		    const ObjErrHandler_t& _eh, 
		    const LogFunc_t&       _log_func = DummyLogger, 
		    const SimpleHandler_t& _sh       = SimpleHandler_t(DummySimpleHandler), 
		    const ErrorHandler_t&  _erh      = ErrorHandler_t(DummyErrorHandler)
		   );
	/// Unregistrates event handlers associated with the object
	~Application();
	/// Creates application
	void Create();
        /// Deletes application
	void Delete();	
	/// Returns state
	bool IsCreated() const {return isCreated;}
	/// Returns application name
	std::string Name() const {return name;}
	/*!
	  \brief Connects to partner 
	  
	  \param _name Partner's name
	  \param _h    Callback
	*/
	void Connect(const std::string& _name, const AppStreamCreated_t& _h = DummyAppStreamCreatedHandler);
	/*!
	  \brief Sets "on stream created" callback
	  
	  \param _h    Callback
	*/	
        void SetCbk(const AppStreamCreated_t& _h);
private:
	/// Identifiers of regular expressions
	typedef enum{genericRE, onCreateRE, onDeleteRE,onStreamRE,onReceivedRE,onReadRE,onConnectedRE} RegExpTypes_t;
	/// Map <<Regular expression ID>> - <<Regular expression>>
	typedef std::map<RegExpTypes_t, std::string> RegExps_t;
	/// Map <<Regular expression ID>> - <<Listener id>>
	typedef std::map<RegExpTypes_t, unsigned> RegExpsIds_t;
	/// Map <<Stream ID>> - <<Stream>>
	typedef std::map<const std::string, Loki::SmartPtr<AppStream> > StreamsMap_t;	
public:
        /*!
           \typedef const_iterator

           \brief Defines const_iterator via container's const_iterator
        */
	typedef StreamsMap_t::const_iterator const_iterator;
        /*!
           \brief Returns first element of container
        */	
	const_iterator begin() const {return streams_.begin();}
        /*!
           \brief Returns "last + 1" element of container
        */		
	const_iterator end() const {return streams_.end();}
        /*!
           \brief Returns an iterator to key, or an iterator to the end of the container if key is not found.
        */		
	const_iterator find(const std::string& key) const {return streams_.find(key);}
        /*!
           \brief Returns the number of elements in container.
        */		
	unsigned size() const {return streams_.size();}
private:
	/// Object-specific initialization.  Called from constructors 
	void Init();
	/// On create slot
	void onCreate(const boost::cmatch& _event);
	/// On delete slot
	void onDelete(const boost::cmatch& _event);
	/// On new stream slot
	void onStream(const boost::cmatch& _event);
	/// On "RECEIVED ..." slot
	void onReceived(const boost::cmatch& _event);
	/// On "READ" slot
	void onRead(const boost::cmatch& _event);
	/// On "RECEIVED" slot
	void onStreamEmpty(const boost::cmatch& _event);
	/// On "CONNECTED" slot
	void onConnected(const boost::cmatch& _event);
	/// Reference to the Skype Connection
	SkypeConnection& sc;
	/// Application name
	const std::string name;
	/// True, if application was created via Skype API
	bool isCreated;
	/// RegExp name <-> RegEx map
	RegExps_t regExps_;
	/// RegExp name <-> ID map
	RegExpsIds_t regExpsIds_;
	/// "OnStreamCreated" Callback 
	AppStreamCreated_t callback_;
	/// Streams
	StreamsMap_t streams_;
	/// True, if stream is empty
	bool isStreamEmpty;
	/// True, if connected
	bool isConnected;
	///
	ErrorHandler_t errorHandler;
	///
	SimpleHandler_t simpleHandler;
};
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  \class AppFlush

  \brief This class represents application sending modes.
*/
struct AppFlush : public LogAbility{
public:
	/// Enumeration of possible modes
	typedef enum { Datagram, Stream } SendMode_t;
	/// Constructs AppFlush object with "Stream" sending mode
	AppFlush(SendMode_t _mode = Stream);
	/// Returns sending mode
	const SendMode_t& GetMode() const;
private:
	/// Sending mode
	SendMode_t mode;	
};
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*!
  \class AppStream

  \brief This class represents application stream.
*/
class AppStream : public LogAbility {
public:
	/*!
	  \brief Creates application stream object
  
	  \param _app       - Application object
	  \param _stream_id - Stream identifier
	  \param _log_func  - Reference to the logger. This logger will be used to log any messages inside the object.
	*/
	AppStream(Application& _app, const std::string& _stream_id, const LogFunc_t& _log_func = DummyLogger);
	/*!
	  \brief The << operator inserts data into the stream.
  
	  \param _buf - Buffer
	  \return Itself 
	*/
	AppStream& operator<<(const char* _buf);
	/*!
	  \brief The << operator inserts data into the stream.
  
	  \param _buf - Buffer
	  \return Itself 
	*/
	AppStream& operator<<(const std::string& _buf);
	/*!
	  \brief Sends collected buffer (via flush) in accordance with AppFlush's state.
  
	  \param _f - AppFlush object (sending mode)
	  \return Itself 
	*/
	AppStream& operator<<(const struct AppFlush& _f);
	/*!
	  \brief Sends collected buffer immediately.

	  \param _mode - Sending mode

	  \return Itself 
	*/
	AppStream& flush(const AppFlush::SendMode_t& _mode = AppFlush::Stream);
	/*!
	  \brief Returns stream's application object

	  \return Application object
	*/
	Application& GetApp();
	/*!
	  \brief Checks for read ability 

	  \return True if there is data in stream; otherwise false.
	*/
	bool CanRead() const;
	/*!
	  \brief Reads data from stream

	         Should be called if and only if CanRead returns true
	  	 
	  \param _buf - Buffer
	  \return Itself 
	*/	
	AppStream& operator>>(std::string& _buf);
	/*!
	  \brief Appends internal input buffer
	  	 
	  \param _buf - Buffer
	*/	
        void append(const std::string& _buf);
	/*!
	  \brief Appends internal input buffer
	  	 
	  \param _buf - Buffer
	*/	
        void append(const char* _buf);
	/*!
	  \brief Returns stream's name (ID)
	*/	
	std::string Name() const {return name;}
private:
	/// Used SkypeConnection
	SkypeConnection& sc;
	/// Application which owns this stream
	Application& app;
	///
	const std::string name;
	///
	bool canRead;
	/// Input buffer 
	std::queue<std::string> inBuf;
	/// Output buffer
	std::queue<std::string> outBuf;
};
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}; // namespace Ice
#endif
