/*!
  \file Chat.h

  \brief Chat-related stuff, declarations

  $Id: Chat.h,v 1.13 2006/06/13 12:12:35 tarle Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   18.08.2005 20:36:11 MSD
*/

#ifndef SK_CHAT_H
#define SK_CHAT_H

#include "SkypeObj.h"

#include <string>

namespace Ice {

/// Attributes of ChatMsg object
typedef enum {
	cmTimeStamp,
	cmFromHandle,
	cmFromDispName,
	cmType,
	cmStatus,
	cmLeaveReason,
	cmBody,
	cmChatName,
	cmUsers,
	cmSeen
} ChatMsgAttrs_t;

/// ChatMsg attributes description
typedef AttrsDescription<ChatMsgAttrs_t> ChatMsgAttrsDesc_t;

/// Describes atribute of Skype object "Chat Message"
template <>
class AttrsDescription<ChatMsgAttrs_t> : public AttrsDescriptionBase<ChatMsgAttrs_t> {
public:
	/// Fills description
	AttrsDescription();
};

// Forward declaration of ChatMsg object
class ChatMsg;

/// Watchers for the ChatMsg object
typedef StateWatchers<ChatMsg, ChatMsgAttrs_t> ChatMsgWatchers_t;

/// This calss represents Chat Message Skype object
class ChatMsg : public SkypeObject<ChatMsg, ChatMsgAttrs_t, SkypeIdObjCmdFormatter> {
public:
    /*!
      \brief	Constructs ChatMsg object with DEFAULT error handler.  
  		  
	  	Default error handler doesn't handle any error at all.

      \param _sc	Reference to Skype Connection
      \param _id	Call identifier
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */
    ChatMsg(SkypeConnection& _sc, const int& _id, const LogFunc_t& _log_func = DummyLogger); 
     /*!
      \brief	Constructs ChatMsg object with NON-DEFAULT (user-supplied) error handler.  
  		  
      \param _sc	Reference to Skype Connection
      \param _id	Call identifier
      \param _eh	Error hanlder.  This functor will be called in case of errors.
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */
    ChatMsg(SkypeConnection& _sc, const int& _id, const ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger);
    /*!
      \brief Checks if object is alive or not. 

      \return alive if object is alive, finished if not alive and unknown if object doesn't held "status" attribute		
    */  
    LiveMode_t IsAlive() const;
private:
    /// Object-specific initialization.  Called from constructors 
    void Init();	
};


/// Possible types of users search
typedef enum {
	scmChatMsgs,
	scmMissedChatMsgs
} ChatMsgSearchTypes_t;

/// Description of the different users search types
typedef AttrsDescription<ChatMsgSearchTypes_t> ChatMsgSearchDesc_t;

/// Describes users search types
template <>
class AttrsDescription<ChatMsgSearchTypes_t> : public AttrsDescriptionBase<ChatMsgSearchTypes_t> {
public:
	/// Type of attributes of the oobjects returned by search procedure
	typedef ChatMsgAttrs_t ObjAttrs_t; 
	/// Type of objects returned by the search procedure
	typedef ChatMsg ObjType_t;
	/// Fills description
	AttrsDescription();
};



class ChatMsgManager : public SkypeObjManager<ChatMsg, ChatMsgAttrs_t, ChatMsgManager, ChatMsgSearchTypes_t, Searcher> {
public:
    /*!
      \brief	Constructs Chat Message manager with DEFAULT error handler.  
  		  
	  	Default error handler doesn't handle any error at all.

      \param _sc	Reference to Skype Connection
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */	
	ChatMsgManager(SkypeConnection& _sc, const LogFunc_t& _log_func = DummyLogger); 
     /*!
      \brief	Constructs Chat Message manager with NON-DEFAULT (user-supplied) error handler.  
  		  
      \param _sc	Reference to Skype Connection
      \param _eh	Error hanlder.  This functor will be called in case of errors.
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */
	ChatMsgManager(SkypeConnection& _sc, const ChatMsg::ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger); 
	/*!
	  \brief Parses event related to chat messages
  
	  \param _match Parsed regular expression
	  \return message identifier
	*/
	int ParseEvent(const boost::cmatch& _match);
private:
	/// Connection to Skype
	SkypeConnection& sc;
	void Init();
};

/// Attributes of Chat object
typedef enum {
	chName,
	chTimeStamp,
	chAdder,
	chStatus,
	chPosters,
	chMembers,
	chTopic,
	chChatMessages,
	chActiveMembers,
	chFriendlyName,
	chRecentChatMessages
} ChatAttrs_t;

/// Chat attributes description
typedef AttrsDescription<ChatAttrs_t> ChatAttrsDesc_t;

/// Describes atribute of Skype object "Chat"
template <>
class AttrsDescription<ChatAttrs_t> : public AttrsDescriptionBase<ChatAttrs_t> {
public:
	/// Fills description
	AttrsDescription();
};

// Forward declaration of Chat object
class Chat;

/// Watchers for the Chat object
typedef StateWatchers<Chat, ChatAttrs_t> ChatWatchers_t;

/// This calss represents Chat Skype object
class Chat : public SkypeObject<Chat, ChatAttrs_t, SkypeIdObjCmdFormatter, SkypeObjRegExParser, std::string> {
public:
    /*!
      \brief	Constructs Chat object with DEFAULT error handler.  
  		  
	  	Default error handler doesn't handle any error at all.

      \param _sc	Reference to Skype Connection
      \param _id	Chat identifier
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */
    Chat(SkypeConnection& _sc, const std::string& _id, const LogFunc_t& _log_func = DummyLogger); 
     /*!
      \brief	Constructs Chat object with NON-DEFAULT (user-supplied) error handler.  
  		  
      \param _sc	Reference to Skype Connection
      \param _id	Call identifier
      \param _eh	Error hanlder.  This functor will be called in case of errors.
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */
    Chat(SkypeConnection& _sc, const std::string& _id, const ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger);
    /*!
      \brief Checks if object is alive or not. 

      \return alive if object is alive, finished if not alive and unknown if object doesn't held "status" attribute		
    */  
    LiveMode_t IsAlive() const;
private:
    /// Object-specific initialization.  Called from constructors 
    void Init();	
};

/// Possible types of chats search
typedef enum {
	srChats,
	srActiveChats,
	srMissedChats,
	srRecentChats,
	srBookmarkedChats
} ChatSearchTypes_t;

/// Description of the different chats search types
typedef AttrsDescription<ChatSearchTypes_t> ChatSearchDesc_t;

/// Describes chats search types
template <>
class AttrsDescription<ChatSearchTypes_t> : public AttrsDescriptionBase<ChatSearchTypes_t> {
public:
	/// Type of attributes of the oobjects returned by search procedure
	typedef ChatAttrs_t ObjAttrs_t; 
	/// Type of objects returned by the search procedure
	typedef Chat ObjType_t;
	/// Fills description
	AttrsDescription();
};

class ChatManager : public SkypeObjManager<Chat, ChatAttrs_t, ChatManager, ChatSearchTypes_t, Searcher> {
public:
    /*!
      \brief	Constructs Chat manager with DEFAULT error handler.  
  		  
	  	Default error handler doesn't handle any error at all.

      \param _sc	Reference to Skype Connection
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */	
	ChatManager(SkypeConnection& _sc, const LogFunc_t& _log_func = DummyLogger); 
     /*!
      \brief	Constructs Chat manager object with NON-DEFAULT (user-supplied) error handler.  
  		  
      \param _sc	Reference to Skype Connection
      \param _eh	Error hanlder.  This functor will be called in case of errors.
      \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
    */	
	ChatManager(SkypeConnection& _sc, const Chat::ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger); 
	/*!
	  \brief Parses event related to chat
  
	  \param _match Parsed regular expression
	  \return chat identifier
	*/
	std::string ParseEvent(const boost::cmatch& _match);
	/*!
	  \brief Initiates a chat with remote subscriber(s)
  	  
	  \param _name Remote subscriber(s) name(s) (skype name(s)). If the number of subscribers
		       more than 1, skype names should be delimeted by comma (',')
	*/
	void MakeChat(const std::string& _name);
private:
	/// Connection to Skype
	SkypeConnection& sc;
	void Init();
};


}; // namespace Ice
#endif // SK_CHAT_H
