/*!
  \file h/LowLevel.h

  \brief Low-level ++Skype API declarations

  $Id: LowLevel.h,v 1.23 2006/06/12 14:34:53 vit Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   19.07.2005 23:14:03 MSD
*/

#ifndef LOWLEVEL_H
#define LOWLEVEL_H

#include "Cfg.h"
#include "Loggers.h"

#include <loki/Functor.h>
#include <loki/SmartPtr.h>
#include <loki/static_check.h>

#include <string>
#include <sstream>

/// Low-level skype event.  
typedef Loki::Functor<void, LOKI_TYPELIST_1(const std::string&)> RawNotify_t;

/*!
  \brief Application -> Skype commands formatter

  This template should be mixed to the inheritance tree of the #RawSkypeConnection interface
  implementation.  It implements operator<< for most useful types in application -> Skype communications:
  string, integer and boolean. 

  To send data to the skype this template rans void Host::flush(const std::&string) method. The argument is the string
  to be flushed.
*/  
template <class Host> class SkypeCmdFormatter : public Numerator<int> {
public:
	/// Reference to host class is required to run host's method flush()
	SkypeCmdFormatter(Host& _host);
	/// Enables Skype commands numeration
	void EnableIds();
	/// Disables Skype commands numeration
	void DisableIds();
	/// Checks the status of commands numeration mechanism: is it turned ON or OFF
	bool IsIdsEnabled() const;
	/// "<< flush", "<< EnableIds", "<< DisableIds" syntax
	SkypeCmdFormatter& operator<<(SkypeCmdFormatter& (*_f)(SkypeCmdFormatter&));
	/// Constant string parts
	SkypeCmdFormatter& operator<<(const char* _s);
	/// String parts
	SkypeCmdFormatter& operator<<(const std::string& _s);
	/// Booleans are sent to Skype as "ON" (for true) and "OFF" (for false)
	SkypeCmdFormatter& operator<<(const bool& _b);
	/// Integer pars
	SkypeCmdFormatter& operator<<(const int& _i);
	/// Flushed output.  Calls host's flush function to do it
	SkypeCmdFormatter& flush();
private:	
	/// Reference to host class
	Host& host;
	/// Skype command.  It is composed by "<<" operators and is cleared inside flush()
	std::string skype_cmd;
	/// Is command numeration functionality enabled?
	bool numeration_enabled;
};

/// This template function is required for correct implementation of "RawSkypeConnection << flush" syntax
template <class Host> SkypeCmdFormatter<Host>& flush(SkypeCmdFormatter<Host>& s);

/// This template function is required for correct implementation of "RawSkypeConnection << EnableIds" syntax
template <class Host> SkypeCmdFormatter<Host>& EnableIds(SkypeCmdFormatter<Host>& s);

/// This template function is required for correct implementation of "RawSkypeConnection << DisableIds" syntax
template <class Host> SkypeCmdFormatter<Host>& DisableIds(SkypeCmdFormatter<Host>& s);

/*!
  \brief Public interface for the low-level connections to skype.

  Implementations should specialize this template for operation system represented by template parameter T.
  It is not recommended to use this interface directly.  Prefer SkypeConnection instead.  
*/  
template <OS_t T = OS> class RawSkypeConnection : public SkypeCmdFormatter< RawSkypeConnection<T> >,
						  public LogAbility 
{
public:
	/*!
	  \brief  Stores functor which will be executed upon messages (Skype -> application) receiving

	  \param _cbk Call-back functor. It will be executed for any message received from Skype
	  \param _app_name Application's name (usually argv[0]).  
	  \param _log_func Logger to use by RawSkypeConnection's functions
	*/  
	RawSkypeConnection(const RawNotify_t& _cbk, const std::string& _app_name, LogFunc_t _log_func = DummyLogger) 
			{ LOKI_STATIC_CHECK(0, YOUR_OS_IS_NOT_SUPPORTED_BY_LIBRARY_SORRY); };	
  	
	/*! 
	   \brief Attach to Skype. Do nothing if already attached. 
	   
	   \throw SkypeDisconnect Error occured during connect.		
	   \throw SkypeAccessDenied Skype user doesn't allow this application to connect to Skype
	*/ 
	void Attach() { LOKI_STATIC_CHECK(0, YOUR_OS_IS_NOT_SUPPORTED_BY_LIBRARY_SORRY); };
	/// Returns true if attached to Skype, false - otherwise
	bool IsAttached(void) { LOKI_STATIC_CHECK(0, YOUR_OS_IS_NOT_SUPPORTED_BY_LIBRARY_SORRY); return false; };
	/// Detach from Skype.  Do nothing if already detached
	void Detach() { LOKI_STATIC_CHECK(0, YOUR_OS_IS_NOT_SUPPORTED_BY_LIBRARY_SORRY); };
	/// Send message to Skype
	void flush(const std::string&) { LOKI_STATIC_CHECK(0, YOUR_OS_IS_NOT_SUPPORTED_BY_LIBRARY_SORRY); };
	/*!
	  \brief Read (if any) and process messages from Skype (Skype notifications and command responses)
	  
	  See implementation-specific documentation for details
	*/  
	bool operator()() { LOKI_STATIC_CHECK(0, YOUR_OS_IS_NOT_SUPPORTED_BY_LIBRARY_SORRY); return false; }; 
};

#ifdef USE_PROXY

namespace Impl {
/// Helper class to implement RawSkypeConnection<os_proxy>
class ProxyLowLevelHelper {
public:
	virtual ~ProxyLowLevelHelper() {};
};
}; // namespace Impl

/// RawSkypeConnection specialization for network proxy
template <> class RawSkypeConnection<os_proxy> : public SkypeCmdFormatter< RawSkypeConnection<os_proxy> >,
					         public LogAbility 
{
public:
	/*!
	  \brief Stores functor which will be executed upon messages (Skype -> application) receiving

	  \param _cbk Call-back functor. It will be executed for any message received from Skype
	  \param _app_name
	  \param _log_func Logger to use inside RawSkypeConnection 
	*/  
	RawSkypeConnection(const RawNotify_t& _cbk, const std::string& _app_name, LogFunc_t _log_func = DummyLogger);	
	void Attach(const std::string& desc_);
	/// Returns true if attached to Skype, false - otherwise
	bool IsAttached(void) { return attach_result; };
	/// Detach from Skype. Do nothing if already detached
	void Detach();
	/*!
	  \brief Send mesage to Skype via TCP 

	  \param _cmd Skype command

	  \throw SkypeDisconnect Disconnected from Skype 
	*/  
	void flush(const std::string& _cmd);
	/*!
	  \brief	Read (if any) and process messages from Skype (Skype notifications and 
	  		command responses).

	  		This is a very simple method. It checks for new messages in queue and
			runs user-defined call-back functor (see RawSkypeConnection constructor) 
			for each received message. See also Attach docs for some implementation 
			details.
			
	  \throw AssertionViolation  If method is executed when connection to D-BUS bus is not established.
	  			     It is thrown only if debug mode is on (DbgMode is set to debug 
				     in Cfg.h).
			
	  \return	True if at least 1 message was received.  False otherwise.
	*/  
	bool operator()();
private:
	/// Call-back functor.  
	RawNotify_t cbk;
	/// The value of timeout for select (in microseconds) if dont_select is false
	unsigned timeout;
	/// Is connection established?
	bool attach_result;
	/// Proxy implementation - socket and protocol fields
	Loki::SmartPtr<Impl::ProxyLowLevelHelper> impl;
};
#endif // USE_PROXY

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/
template <class Host> 
inline SkypeCmdFormatter<Host>::SkypeCmdFormatter(Host& _host)
			       : host(_host), numeration_enabled(true)
{};

template <class Host> 
inline void SkypeCmdFormatter<Host>::EnableIds()
{ 
  numeration_enabled = true;
};

template <class Host> 
inline void SkypeCmdFormatter<Host>::DisableIds() 
{
  numeration_enabled = false;
};

template <class Host> 
inline bool SkypeCmdFormatter<Host>::IsIdsEnabled() const
{
  return numeration_enabled;	
};

template <class Host>
SkypeCmdFormatter<Host>& SkypeCmdFormatter<Host>::operator<<(const char* _s)
{
  skype_cmd.append(_s);
  return *this;
};

template <class Host> 
SkypeCmdFormatter<Host>& SkypeCmdFormatter<Host>::operator<<(const std::string& _s)
{
  skype_cmd.append(_s);
  return *this;
};

template <class Host> 
SkypeCmdFormatter<Host>& SkypeCmdFormatter<Host>::operator<<(const bool& _b)
{
  if(_b) skype_cmd.append("ON");
  else skype_cmd.append("OFF");	 
  
  return *this;
};

template <class Host> 
SkypeCmdFormatter<Host>& SkypeCmdFormatter<Host>::operator<<(const int& _i)
{
  std::ostringstream ost; 
  ost << _i;
  skype_cmd.append(ost.str());
  
  return *this;  	
};

template <class Host> 
inline SkypeCmdFormatter<Host>&  SkypeCmdFormatter<Host>::operator<<(SkypeCmdFormatter& (*_f)(SkypeCmdFormatter&))
{
  return _f(*this);
};

template <class Host> 
inline SkypeCmdFormatter<Host>& SkypeCmdFormatter<Host>:: flush()
{
//  const char* n = __FUNCTION__;
  
  if(numeration_enabled) {
	std::ostringstream ost;
	ost << '#' << IncrNum() << " " << skype_cmd;
//	fprintf(stderr, "%s (numeration enabled): Sending '%s' to Skype\n", n, ost.str().c_str());
	host.flush(ost.str());
  } else {
//	fprintf(stderr, "%s (numeration disabled): Sending '%s' to Skype\n", n, skype_cmd.c_str());
	host.flush(skype_cmd);
  };

  skype_cmd.clear();
  return *this;	
};	

template <class Host> 
inline SkypeCmdFormatter<Host>& flush(SkypeCmdFormatter<Host>& s)
{ 
  return s.flush();
};

template <class Host> 
inline SkypeCmdFormatter<Host>& EnableIds(SkypeCmdFormatter<Host>& s)
{
  s.EnableIds();
  return s;
};

template <class Host> 
inline SkypeCmdFormatter<Host>& DisableIds(SkypeCmdFormatter<Host>& s)
{
  s.DisableIds();
  return s;
};

#include "unix/LowLevel.h"
#include "win/LowLevel.h"

#endif // LOWLEVEL_H
