/*!
  \file SkypeObj.h

  \brief Skype object & object manager declarations

  $Id: SkypeObj.h,v 1.53 2006/06/17 15:18:15 vit Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   07.08.2005 10:21:56 MSD
*/

#ifndef SKYPE_OBJ_H
#define SKYPE_OBJ_H

#include "ObjCmdFormatters.h"
#include "Search.h"

#include <string>
#include <sstream>
#include <map>
#include <set>
#include <algorithm>
#include <functional>
#include <vector>

#include <loki/Functor.h>
#include <loki/SmartPtr.h>
#include <loki/TypeTraits.h>

#include <boost/regex.hpp>

namespace Ice {

/*!
  \brief Skype object state watcher.

  Watches the state of skype object and calls functor associated with attribute value change event.

  \param Host 		Watched SkypeObject
  \param Attrs 		Attrbute type (usually some kind of enum type)
*/  
template <class Host, typename Attrs>
class StateWatchers {
public:
	/// The set of Skype object's attributes.
	typedef std::set<Attrs> AttrSet_t;
	/// Initialization
	StateWatchers() : do_watch_in_progress(false) {};
	/*!
	  \brief Callback functor executed where the value of watched attribute changes

	  Parameters:
	    - Reference to watched SkypeObject
	    - Changed attribute
	    - New value of changed attribute
	*/	
	typedef Loki::Functor<void, LOKI_TYPELIST_3(Host&, const Attrs&, const std::string&)> ObjChanged_t;
	/*!
  	  \brief Registrates new watcher  

	  \param _attrs  The set of watched Skype object's attributes 
	  \param _cbk	 Callback functor which should be executed when the value of any attribute from the set 
	  		 _attrs changes
	  \param _not_in Mode of _attrs checking: if false, checks if attribute is included in _attrs, 
	 		 else checks if it is NOT included in _attrs.

	  \return Identifier of registered watcher
	*/
	unsigned RegisterWatcher(const AttrSet_t& _attrs, const ObjChanged_t& _cbk, const bool& _not_in = false);
	/*!
	  \brief Unregistrates watcher previously registered with #RegisterWatcher
  
	  	 If there is no registered watcher with id equal to _id, nothing happens
	  
	  \param _id  Watcher identifier returned by call to #RegisterWatcher
	*/
	void UnRegisterWatcher(const unsigned& _id);
	/// Unregistrates all previously defined with #RegisterWatcher watchers
	void clear();
	/*!
	  \brief Executes all registered watchers. 
  		 
	  	 If _attr is included in the atribute set of inspected watcher, registered calback functor is executed. 
		 
	  \param _host  Reference to watched Skype object
	  \param _attr	Changed attribute name	
	  \param _value New value of the attribute _attr
	*/
	void operator()(Host& _host, const Attrs& _attr, const std::string& _value);
private:
	/// Watcher's data required for registration	
	class WatcherDetails {
	public:	
		/*!
	  	  \brief WatcherDetails creation	
	  
	  	  \param _attrs  The set of watched Skype object's attributes 
	  	  \param _cbk	 Callback functor which should be executed when the value of any attribute from the set 
		  		 _attrs changes
	  	  \param _not_in Mode of _attrs checking: if false, checks if attribute is included in _attrs, 
		  		 else checks if it is NOT included in _attrs.

		*/
		WatcherDetails(const AttrSet_t& _attrs, const ObjChanged_t& _cbk, const bool& _not_in = false);
	private:	
		/// Watched attributes
		AttrSet_t attrs;
		/// Callback
		ObjChanged_t callback;
		/// Check mode
		bool not_in;
		/// StateWatchers::DoWatch needs the access to private members of this class
		friend 
		#ifdef SKYPE_MSC
		typename
		#else
		class
		#endif // SKYPE_MSC 
		StateWatchers<Host, Attrs>::DoWatch;
	};	
	/// Pointer to #Ice::StateWatchers::WatcherDetails
	typedef Loki::SmartPtr<WatcherDetails> WatcherDetailsPtr_t;
	/// This functor is used to check all watchers and call associated functions if it is needed
	class DoWatch {
	public:
		/*!
		  \brief  Stores values which are required inside operator()

		  \param _host Reference to host object
		  \param _attr  Changed attribute
		  \param _value New attribute value
		*/
		DoWatch(Host& _host, const Attrs& _attr, const std::string& _value);
		/*!
		  \brief This function is called by Registrator::for_each method
	  
	 	  \param _id		Watcher identifier
		  \param _watcher	Pointer to watcher's details
		*/
		void operator()(const unsigned& _id, WatcherDetailsPtr_t& _watcher);
	private:
		/// Reference to the host object
		Host& host;
		/// Changed attribute
		const Attrs& attr;
		/// New value of the attribute
		const std::string& value;
	};	
	bool do_watch_in_progress;
	/// Watchers to remove (post-processing.  For watchers removed in watch callback function)
	std::vector<unsigned> watchers_to_remove;
	/// Watcher's details
	Registrator<WatcherDetailsPtr_t> watchers;
};

/*!
  \defgroup SkypeObjParser SkypeObjParser policy of Ice::SkypeObject template.
  
  	    This policy is responsible for parsing responses and events associated with the object.
	    See documentation on #Ice::SkypeObject for details.
*/

/*!
  \brief SkypeObjParser policy implementation based on regular expressions. 
  
  Parses events and responses with the help of regular expressions. Object of this class is the part of #Ice::SkypeObject
  instance.  
  
  \param Attrs The type of skype object's attributes which are parsed
  
  \ingroup SkypeObjParser
*/  
template <typename Attrs> 
class SkypeObjRegExParser {
public:
	/// Dummy constructor is required by compiler
	SkypeObjRegExParser() {};
	/*!
	  \brief  Response parser

	  \param _resp Response string received from Skype

	  \return Attribute value 
	*/
	const std::string ParseResponse(const std::string& _resp);
	/// Attribute <-> its value pair
	typedef std::pair<Attrs, std::string> AttrValuePair_t;
	/*!
	  \brief Event parser
  
 	  \param _event Event received from Skype and parsed by #SkypeConnection

	  \return  Attribute name and value pair
	*/
	const AttrValuePair_t& ParseEvent(const boost::cmatch& _event);
	/// Structure for storing the parameters of #SkypeObjRegExParser.  
	/// It is required for fine tuning of #SkypeObjRegExParser.
	struct SkypeObjRegEx {
		/// Regular expression for Skype response parsing
		boost::regex resp;
		/// The oredering number of regular expression resp submatch where attribute value is contained.
		unsigned char resp_value_pos;
		/// The ordering number of regular expression event submatch where attribute name is contained.
		unsigned char event_attr_pos;
	/// The ordering number of regular expression event submatch where attribute value is contained.
		unsigned char event_value_pos;
	};
	/*!
	  \brief  Set regular expression used by the parser.  Users should call this method before any other methods.

	  \param _regex Regular expressions associated with this parser
	*/
	void SetRegEx(const struct SkypeObjRegEx& _regex);
private:
	/// Regular expressions used by this parser
	SkypeObjRegEx re;
	/// Object atrtributes description
	const AttrsDescription<Attrs> attrs_desc;
	/// regular expression match.  Used by #ParseResponse
	boost::cmatch match;
	/// Result of event parsing.  Used by #ParseEvent 
	AttrValuePair_t parse_event_result;
};

/*!
  \brief A Skype object template.  This class is used as a base by the variety of classes which represents Skype objects. 

	 Features:
	   - Stores the object's attributes. They can be accessed via "[]" operator or constant iterators. Ietrating through
	     attribute set can be done with the help of constant iterators.
	   - Parses Skype events associated with the object. Sends GET/SET-requests to the Skype to get/set the attribute
	     value.  Error handling is supported. 
	   - Watchers mechanism support. 
	  
	   There are two policies introduced by this template: \ref SkypeObjCmdFormatter "SkypeObjCmdFormatter" and 
	   \ref SkypeObjParser "SkypeObjParser".  
	   
	   \e SkypeObjCmdFormatter is responsible for formatting GET/SET command.  These commands are sent to the Skype in order
	   to get or set the value of attribute.  This policy expected to be implemented as template with 2 parameters - the
	   first is the type of Skype object's attributes and second is the type of Skype object itself. Let us call this 
	   template SkypeObjCmdFormatterTestImpl<typename Attrs, class SkypeObj>.  It's constructor should have exactly 2
	   parameters: the reference to Skype object and the reference to SkypeConnection object.  So, constructor declaration
	   looks like this:

 \code
  SkypeObjCmdFormatterTestImpl(SkypeObj& _host, SkypeConnection& _sc); 
 \endcode 
	   
 	   Two member functions are required to be defined (in the public section):
  \code	   
   void FormatGetCmd(const Attrs& _attr);
   template <typename T> void FormatSetCmd(const Attrs& _attr, const T& _value);
  \endcode

  	   These functions should send GET/SET commands to the #SkypeConnection WITHOUT Flush or flush call.
	   This policy implementation may call any constant Skype object's functions.  Most frequently 
	   used functions are #GetId() and #GetName().
	   
	   See templates #Ice::SkypeIdObjCmdFormatter and #Ice::SkypeNoIdObjCmdFormatter declarations for the examples of 
	   SkypeObjCmdFormatter policy implementations.
	   
	   \e SkypeObjParser is responsible for parsing responses and events associated with the object.
	   This policy expected to be implemented as template with one parameter - the type of Skype object's
	   attributes. It should have default constructor (without any parameters).  The following member functions  
	   are required to be defined (in the public section):
 
 \code	   
  const std::string ParseResponse(const std::string& _resp);
  const std::pair<Attr, std::string>& ParseEvent(const boost::cmatch& _event);
 \endcode
 	  
 	   First function is called by the SkypeObject instance when response is receive from Skype to the command previously
	   sent to the Skype.  This function should extract the attribute value from _resp string and return it to the caller.
	   Second function should extract attribute name and value pair from Skype event (preprocessed by #SkypeConnection) and 
	   return them to the caller.  This function is called by #ProcessEvent method of the SkypeObject template. See 
	   documentation on #ProcessEvent for additional details.
  
  \param Host		Concrete implementation of SkypeObject
  \param Attrs		The type of concrete Skype object attributes
  \param SkypeObjCmdFormatter 	SkypeObjCmdFormatter policy,  See above for details
  \param SkypeObjParser		SkypeObjParser policy.  See above for details
  \param Id_t 		The type of object's identifier
*/  
template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser = SkypeObjRegExParser,
		typename Id_t = int
	   >  
class SkypeObject : public LogAbility {
public:
	/*!
	  \brief Skype object error handler

	  Functors of this type are used as error handlers.  They are called when Skype error occurs during GET- or SET-
	  commands.  These functors have the following parameters (from left to right):

	  - Reference to the Skype object where error have occured
	  - Attribute name of the Skype object where error have occured (Last GET/SET command attribute parameter)
	  - Integer Skype error code
	  - Human readable Skype error message
	  - Unmodified Skype error messaged, received from the Skype
	*/  
	typedef Loki::Functor<void, LOKI_TYPELIST_5(Host&, const Attrs&, const int&, const std::string&, const std::string&)> 
														ObjErrHandler_t;
	/// Identifier's type
	typedef Id_t ID_t;
	/// Constant iterator.  Use it to iterate through attributes, stored in this object
	typedef typename std::map<Attrs, std::string>::const_iterator const_iterator;
	/// Attribute description type
	typedef AttrsDescription<Attrs> AttrsDescription_t;
	/// Mode of object presence: alive, not alive, unknown 
	typedef enum { alive, finished, unknown } LiveMode_t;
	/*!
	  \brief Create object with DEFAULT error handler
  
	  The default error handling policy is to log the error with level LlError and ignore it.
	  
	  \param _sc		Reference to the Skype connection
	  \param _host		Reference to the host (usually child class object) of this instance
	  \param _obj_name	Skype name of the object (CALL, CHAT, CHATMESSAGE and so on)
	  \param _id		Identifier of this Skype object
	  \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
	*/
	SkypeObject(SkypeConnection& _sc, Host& _host, const std::string& _obj_name, 
		    const Id_t& _id, const LogFunc_t& _log_func = DummyLogger); 
	/*!
	  \brief Create object with NON-DEFAULT error handler

	  \param _sc		Reference to the Skype connection
	  \param _host		Reference to the host (usually child class object) of this instance
	  \param _obj_name	Skype name of the object (CALL, CHAT, CHATMESSAGE and so on)
	  \param _id		Identifier of this Skype object
	  \param _eh		Error hanlder.  This functor will be called in case of errors.
	  \param _log_func	Reference to the logger. This logger will be used to log any messages inside the object.
	*/
	SkypeObject(SkypeConnection& _sc, Host& _host, const std::string& _obj_name, 
		    const Id_t& _id, const ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger);
	/*!
	  \brief Get the value of stored object's attribute. 

	  	 Note: attribute _attr should be stored in the object.  If it is not really stored, memory access violation
		 can occur.
		 
	  \throw AssertionViolation (Debug mode only) If attribute _attr is not really stored in the object.	 
	  
	  \param _attr The attribute which value is requested

	  \return  Skype object's attribute value
	*/
	const std::string& operator[](const Attrs& _attr) const;
	/*!
	  \brief Find the attribute in the object
  
	  \param _attr The attribute which value is requested

	  \return Constant iterator pointed to the attribute <-> value pair (std::pair<Attrs, std::string>) if attribute is 
	          stored in the object, end() otherwise.
	*/
	const_iterator find(const Attrs& _attr) const;
	/// Points to the beginning of stored attributes sequence
	const_iterator begin() const;
	/// Points to the "last + 1" element of stored attributes sequence
	const_iterator end() const;
 	/*!
	  \brief Sends GET- command to Skype in order to discover the object's attribute.  To handle errors, 
	  	 functor, passed to the constructor, is used.
	  	
	  \param _attr Attribute which value is requested from Skype
	*/
	void AskAttr(const Attrs& _attr);
	/*!
	  \brief Sends GET- command to Skype in order to discover the object's attribute. User supplied error handler
	  	 is called in case of any error.
  
	  \param _attr	Attribute which value is requested from Skype
	  \param _eh	User-supplied error handler   
	*/
	void AskAttr(const Attrs& _attr, const ObjErrHandler_t& _eh);
	/*!
	  \brief Sends SET- command to Skype in order to set the value of the object's attribute.  To handle errors, 
	  	 functor, passed to the constructor, is used.

		 \e T is the type of attribute value. Any type which is supported by SkypeCmdFormatter is allowed here.  

	  \param _attr	Attribute which value should be set
	  \param _value New attribute value
	*/
	template <typename T> void AskForSetAttr(const Attrs& _attr, const T& _value);
	/*!
	  \brief Sends SET- command to Skype in order to set the value of the object's attribute.  User supplied error handler
	  	 is called in case of any error.
		
		 \e T is the type of attribute value. Any type which is supported by SkypeCmdFormatter is allowed here.  

	  \param _attr	Attribute which value should be set
	  \param _value New attribute value
	  \param _eh	Type of attribute value. Any type which is supported by SkypeCmdFormatter is allowed here.  
	*/
	template <typename T> void AskForSetAttr(const Attrs& _attr, const T& _value, const ObjErrHandler_t& _eh);
	/// Returns the identifier of the object (See constructor parameters)
	const Id_t& GetId() const;
	/// Returns Skype-name of the object (See constructor parameters)
	const std::string& GetName() const;
	/// Short name for the watchers associated with the object
	typedef StateWatchers<Host, Attrs> Watchers_t;
	/// Return Object state watcher for current object
	Watchers_t& GetWatchers();
	/*!
	  \brief  Turns on/off "ignore watchers" mode

	  	  If object is in this mode, it doesn't check watchers associated with itself
  
	  \param _ignore Turn on (true) or off (false) "ignnore watchers" mode
	*/
	void IgnoreWatchers(bool _ignore = true);
	/*!
	  \brief Gets skype connection used by the object
  
	  \return SkypeConnection object
	*/
	SkypeConnection& GetSkype();
	/*!
	  \brief Process Skype event.
	  
	  	 This function should be called when Skype event associated with the object occurs.  It can be called from
		 Skype object manager or by #SkypeConnection object directly because the signature of this member function
		 is the same as signature of functor type #RegExNotify_t.  In any case, Skype object manager or object itself 
		 should be registered to get Skype events.  Call #SkypeConnection::RegisterCallback to register.
		 
	  \param _event  Skype event pre-parsed by #SkypeConnection.
	*/
	void ProcessEvent(const boost::cmatch& _event);
protected:
	/// Short name for the instance of this template.  Useful in child classes
	typedef SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t> Parent_t;
	/// Short name for the skype events parser type
	typedef SkypeObjParser<Attrs> Parser_t;
	/// Parser used to parse Skype events. You can tune this object inside constructors of the Host class.  
	Parser_t parser;
private:
	/// GET/SET- commands formatter 
	SkypeObjCmdFormatter<Attrs, Host> cmd_formatter;
	/// Description of Skype object's attributes
	const AttrsDescription<Attrs> attrs_desc;
	/*!
	  \brief Stores the attribute <-> value pair inside object.
  
	  	Calls watchers operator() to inform the application about object state change if it is needed.
		
	  \param _attr_value_pair Attrribute and its value
	*/
	void SetAttrValue(const typename Parser_t::AttrValuePair_t& _attr_value_pair);
	/*!
  	  \brief  Default Skype rror handler. It logs the error with level LlError and ignores it.

  	  \param _me		*this
  	  \param _attr		Attribute, which was the cause of error
  	  \param _err_code	Integer Skype error code
  	  \param _err_txt	Human-readable Skype error description
  	  \param _err_msg	Full error message received from Skype before any post-processing by ++Skype library 
	*/
	void DummyErrHandler(Host& _me, const Attrs& _attr, const int& _err_code, 
			     const std::string& _err_txt, const std::string& _err_msg);
	/// Base class for response and error handlers
	class HandlerImpl {
	public:
		/// Stores references to the attribute and host object
		HandlerImpl(const Attrs& _attr, Host& _host);
	protected:
		/// Attribute
		Attrs attr;
		/// Host object
		Host& host;
	};
	/// Error handler
	class ErrHandlerImpl : public HandlerImpl {
	public:
		/// Stores references to real error handler functor, attribute and host object
		ErrHandlerImpl(const ObjErrHandler_t& _eh, const Attrs& _attr, Host& _host);
		/*!
		  \brief Error handler.  It is called by #SkypeConnection when error occurs

		  \param _err_code	Integer Skype error code
		  \param _err_txt	Human-readable Skype error description
		  \param _err_msg	Unmodified Skype error message
		*/
		void operator()(const int& _err_code, const std::string& _err_txt, const std::string& _err_msg); 
	private:
		/// Error handler functor
		ObjErrHandler_t eh;
	};
	/// Response handler
	class RespHandlerImpl : public HandlerImpl {
	public:
		/// Stores references to the attribute and host object
		RespHandlerImpl(const Attrs& _attr, Host& _host);
		/*!
		  \brief Response handler. It is called by #SkypeConnection when response to GET or SET command is 
		  	 received from Skype.

		  \param _resp Response string received from Skype
		*/  
		void operator()(const std::string& _resp);
	};
	/// Object identifier
	const Id_t id;
	/// Reference to the host
	Host& host;
	/// Connection with Skype
	SkypeConnection& sc;
	/// Default (supplied in constructor) error handler. It is used in short AskAttr and AskForSetAttt versions. 
	ObjErrHandler_t eh;
	/// Attributes of Skype object
	std::map<Attrs, std::string> attrs;
	/// Object's name
	const std::string obj_name;
	/// Object's state watchers supplied in constructor
	Watchers_t watchers;
	/// "Ignore watchers" mode
	bool ignore_watchers;
	/// Selector of log
	enum AssertLogSelector {Integral, String}; 
        /// Concrete implementation of LoggAssert for integral types
        template <class T> 
        void LoggImpl (unsigned attr, T obj_id, Loki::Int2Type<Integral>) const
          { 
	     log(LlDbg,"ERROR : Attribute = '%d' Object ID = '%d'",attr,obj_id);
          } 
        /// Concrete implementation of LoggAssert for NON integral types (std::string)
        template <class T> 
        void LoggImpl (unsigned attr, T obj_id, Loki::Int2Type<String>) const
          { 
	     log(LlDbg,"ERROR : Attribute = '%d' Object ID = '%s'",attr,obj_id.c_str());
          } 
        /// LoggAssert calls concrete implementation
        template <class T> 
        void LoggAssert (unsigned attr, T obj_id) const 
          { 
             enum {IdType = Loki::TypeTraits<T>::isStdIntegral ? Integral : String }; 
             LoggImpl(attr, obj_id, Loki::Int2Type<IdType>()); 
          }	
};

/*!
  \brief Manager of Skype objects 

  This template is used as base to construct managers of concrete skype objects, such as CallsManager, VoiceMailManager, 
  UsersManager and so on. Basic object management functionality is implemented in SkypeObjManager template.

  The key features of this template:
  
  - Object creation and storing  
  - Object lifetime management
  - "Listeners" concept support via "Object created" signals (implemented as callback functors)
  - Search interface
  
  Host object should have the following member functions:
  - int ParseEvent(const boost::cmatch& _match). Return value is the identifier of object which should process the event.
    This function is executed when new event is received from Skype.  To add event to the list of handled events, call
    #AddEvent.
  
  This template supports Search interface via SearchPolicy policy.   There are two SearchPolicy policy implementations 
  -- Searcher and NoSearch.  The default SearchPolicy policy is NoSearch.
  
  \param ObjHost	Type of objects controlled by the manager 
  \param ObjAttrs	Attributes of objects controlled by the manager
  \param Host		SkypeManager implementation - child of this template.
  \param SearchTypes    Search policy parameters. See documentation on NoSearch and SearchPolicy templates for details
  \param SearchPolicy	Search policy implementation.  See documentation on NoSearch and SearchPolicy templates for details
*/
template < class ObjHost, typename ObjAttrs, class Host, typename SearchTypes = Loki::NullType, template <class, class> class SearchPolicy = NoSearch > 
class SkypeObjManager : virtual public LogAbility, public SearchPolicy<SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>, SearchTypes> {
public:	
	/*! 
	  \brief "Object created" signal

	  Functor of this type is called upon new Skype object creation.  Its parameter is the reference to that object.
	  SkypeObjManager analyses value returned by the functor. If true was returned, SkypeObjManager adds this 
	  newly created object to the list of "useful" objects.  Useful objects are objects which are needed by the
	  application.  So, these objects can not be deleted by manager in order to optimize occupied memory space.
	  If application is intrested in the object, ObjCreated_t functor should return true.  When the object becomes
	  unintresting to the application, it should call #ReleaseObj method of the manager.

	  Note: Each registrated listener which uses the object should inform the manager about its interest via
	  ObjCreated_t return value and inform about loss of the interest vai #ReleaseObj method.  If the first condition
	  is not met, object can be deleted by the manager at any time and application will not be able to know about this fact.
	  If the second condition is not met, memory leak occurs.
	*/ 
	typedef Loki::Functor<bool, LOKI_TYPELIST_1(ObjHost&)> ObjCreated_t;
	/// The type of object's identifier
	typedef typename ObjHost::ID_t Id_t;
	/// The type of object
	typedef Host HOST_t;
	/// The set of Skype object's attributes 
	typedef std::set<ObjAttrs> AttrSet_t;
	/// Smart pointer to Skype object
	typedef Loki::SmartPtr<ObjHost, Loki::RefCounted, Loki::AllowConversion> ObjPtr_t;
	/*!
  	  \brief  Construction of SkypeObjManager with DEFAULT error handler  

	  	  Default error handler doesn't handle any error at all.

	  \param _sc	Reference to SkypeConnection object
	  \param _host	Reference to host object (derived from this template)
	  \param _log_func Logger functor to use for logging	
	*/
	SkypeObjManager(SkypeConnection& _sc,  Host& _host, const LogFunc_t& _log_func = DummyLogger);
	/*!
  	  \brief  Construction of SkypeObjManager with NON-DEFAULT (user-supplied) error handler  

	  \param _sc	Reference to SkypeConnection object
	  \param _host	Reference to host object (derived from this template)
	  \param _eh	Error handler
	  \param _log_func Logger functor to use for logging	
	*/
	SkypeObjManager(SkypeConnection& _sc, Host& _host,
			const typename ObjHost::ObjErrHandler_t& _eh, const LogFunc_t& _log_func = DummyLogger); 
	/*!
	  \brief  Registrates new listener
  
	  \param _attrs The set of object's attributes required to be set to call _cbk functor
	  \param _cbk	Callback functor.  It will be called upon the following conditions: new Skype object is
			created and all its attributes from the _attrs set are set.  Execution of this callback function
			also called "Object created" signal emiting.  See documentation on #ObjCreated_t for 
			additional details related to the callback functor.

	  \return Listener identifier. 
	*/
	unsigned RegisterListener(const AttrSet_t& _attrs, const ObjCreated_t& _cbk);
	/*!
  	  \brief  Unregistrates listener. 
	  	 
	  If there is no registered listener identified by _id, nothing happens.
	  
	  \param _id Identifier of the listener returned by #RegisterListener.
	*/
	void UnRegisterListener(unsigned _id);
	/*!
	  \brief Find the object by its id

	  \param _id		Object's identifier
	  \param _want_to_own	Is this object intrested to the caller?  If true and object is found, 
	  			it will be added to the list of "useful" objects.  See #ObjCreated_t
				description for details.  If you set _want_to_own to true, it is required
				to release the object with #ReleaseObj method later.

	  \return  Pointer to the object if it was found, 0 otherwise
	*/
	ObjPtr_t find(const Id_t& _id, bool _want_to_own = false);
	/*!
	  \brief This method is used by the application to inform the manager about loss of interest to Skype object _obj.
	  	 
	  	 See documentation on #ObjCreated_t for details.  Note: objects should be acquired by the application via
		 return value of #ObjCreated_t functor and should be released via call to this (#ReleaseObj) method.
  
	  \param _obj Unintrested object
	*/
	void ReleaseObj(const ObjHost& _obj);
	/*!
	  \brief Deletes object from list of released objects 
  
	  \param _id Object identifier
	*/	
	void UnReleaseObj(const Id_t& _id);
	/*!
	  \brief Free memory allocated for released objects

	  This procedure should be called on regular basis.  It should NOT be called from one of ++skype callback functors.
	*/  
	void Squeeze();
	
	/*!
	  \brief Create skype object by ID directly
	*/
	ObjHost& Create(Id_t _id);
	/*!
	  \brief Returns number of objects
	  	 
	*/	
	unsigned Size() const;
protected:
	/*!
	  \brief  Add event to the list of events handled by object.  

	  Note: Host should have method ParseEvent to parse received events.  See documentation on #SkypeObjManager
	  for details.
  
	  \param _regex Regular expression.  Should describe an event(s).  It is used in call to 
			#SkypeConnection::RegisterCallback.
	  
	  \throw AssertionViolation (Debug mode only) If _regex can not be compiled. 
	*/
	void AddEvent(const std::string& _regex);
	/// Short name for the instance of this template.  Useful in child classes
	typedef SkypeObjManager< ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy > Parent_t; 
private:
	/*!
	  \brief Processes an event (notify) from Skype

	  \param _match	Regular expression matched variables
	*/
	void EventHandler(const boost::cmatch& _match);
	/// Connection to Skype
	SkypeConnection& sc;
	/// Host object
	Host& host;
	/// Listener's data
	struct Listener {
		/// Simple constructor to fill data fields
		Listener(const AttrSet_t& _attrs, const ObjCreated_t& _cbk, const unsigned& _id);
		/// The set of object's attributes required to be set to emit "ObjCreated" signal to the listener
		AttrSet_t attrs;
		/// "ObjCreated" signal (listener's callback functor)
		ObjCreated_t callback;
		/// Listener id
		unsigned id;
	};
	/// Smart pointer to the Listener's data
	typedef Loki::SmartPtr<Listener> ListenerPtr_t;
	/// This predicate is used to find Listener with specific id inside ListenersLst_t lists
	class ListenerIdEqual : public std::unary_function<ListenerPtr_t, bool> {
	public:
		/*!
		  \brief  Stores id for comparison
		  
		  \param _id Listener identifier
		*/
		ListenerIdEqual(const unsigned& _id);
		/*!
	  	  \brief Compares p->id with id
	  
	  	  \param _p Pointer to Listener structure

		  \return true if p->id equal to id, false otherwise 
		*/
		bool operator()(const ListenerPtr_t& _p) const; 
	private:
		/// Id for comparison
		const unsigned id;
	};
	/// Type to store registered listeners
	typedef std::vector<ListenerPtr_t> ListenersVec_t;
	/// Registered listeners
	ListenersVec_t listeners;
	/// Listeners didn't informed about object creation.  Key: Skype object identifier
	std::map<Id_t, ListenersVec_t> waiting_listeners;
	/*!
	  \brief Removes listeners from the _vec with identifier equal to _id
  
	  \param _vec	Vector to process
	  \param _id	Identifier to compare
	*/
	void RemoveListeners(ListenersVec_t& _vec, const unsigned& _id);
	/*!
	  \brief Updates required_attrs set.  
	  
	  Looks for elements in _attrs which required_attrs doesn't contain and adds these elements to required_attrs

	  \param _attrs	The set of object's attributes 
	*/
//	public:
	void UpdateRequiredAttrs(const AttrSet_t& _attrs);
//	private:
	/*!
	  \brief  Handles information about attribute's values changes.

	  This method is called only for objects which has non-empty list of waiting listeners.

	  \param _obj	Object
	  \param _attr  Changed (created) attribute
	  \param _value Attribute value
	*/
	void ObjChanged(ObjHost& _obj, const ObjAttrs& _attr, const std::string& _value);
	/*!
	  \brief Checks if there are listeners waiting for object creation.  If such listeners are found
	  	 and all required attributes are set, ObjCreated_t signal is emited.
  
	  \param _obj Created object
	*/
	void CheckListeners(ObjHost& _obj);
	/// Skype object atrtributes description
	const AttrsDescription<ObjAttrs> attrs_desc;
	/// Skype objects, controlled by SkypeObjManager.  Key: Skype object identifier
	std::map<Id_t, ObjPtr_t> objects;
	/// This structure stores released objects data.  
	struct ReleasedObjData {
		/// Constructs empty object. It is required by std::map
		ReleasedObjData() :
			tm(boost::posix_time::microsec_clock::universal_time())
		{};
		/*!
	  	  \brief Constructs structure 
	  	  
		  \param _p Pointer to released object
		*/
		ReleasedObjData(const ObjPtr_t& _p);
		/// Pointer to the object
		ObjPtr_t object;
		/// Time of object release
		boost::posix_time::ptime tm;
	};	
	/// Released objects. Events related to these objects are ignored.
	/// This container is cleaned on the regular basis
	std::map<Id_t, ReleasedObjData> released_ids;	
	/// Reference counting.  Key: Skype object ID (the same as in #objects).  Value: reference count for this object
	std::map<Id_t, int> obj_refs;
	/// Numerator is used to assign identifiers to the listeners
	Numerator<unsigned> numerator;
	/// Object's error handler for every controlled object
	Loki::SmartPtr<typename ObjHost::ObjErrHandler_t> object_eh_ptr; 
	/// Attributes, that are required to be asked from Skype after object creation
	AttrSet_t required_attrs;
	/// Watchers for objects controlled my manager
	typedef StateWatchers<ObjHost,ObjAttrs>	ObjWatchers_t;
	/// Identifier of watcher which refers to the #ObjChanged member function
	unsigned obj_watch_id;
	/// Set to true if CheckListeners is running
	bool check_listeners_run;
	/*!
	 *  \brief  Listeners to remove
	 *   
	 *  if #UnRegisterListener is called inside ObjCreated_t callback (if check_listeners_run == true)
	 *  listeners are not removed. Instead, they are put into this vector.  They are removed upon
	 *  finish of the callback function 
	*/
	std::vector<unsigned> listeners_to_remove;
};

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/

//************************************ StateWatchers::WatcherDetails**************************************//
template <class Host, typename Attrs> 
inline StateWatchers<Host, Attrs>::WatcherDetails::WatcherDetails(const AttrSet_t& _attrs, const ObjChanged_t& _cbk, 
								  const bool& _not_in)
			   : attrs(_attrs), callback(_cbk), not_in(_not_in)
{};	

//************************************ StateWatchers::DoWatch*********************************************//
template <class Host, typename Attrs> 
inline StateWatchers<Host, Attrs>::DoWatch::DoWatch(Host& _host, const Attrs& _attr, const std::string& _value)
				  : host(_host), attr(_attr), value(_value)
{};

template <class Host, typename Attrs> 
inline void StateWatchers<Host, Attrs>::DoWatch::operator()(const unsigned& _id, WatcherDetailsPtr_t& _watcher)
{
  bool found = (_watcher->attrs.find(attr) != _watcher->attrs.end());
  bool r =  (_watcher->not_in && (!found)) || ((!_watcher->not_in) && found); // logical XOR
  if(r) _watcher->callback(host, attr, value);	  
}	
	
//************************************ StateWatchers *****************************************************//

/*
template <class Host, typename Attrs> 
unsigned StateWatchers<Host, Attrs>::RegisterWatcher(const WatcherDetailsPtr_t& _details)
{
  return watchers.Register(_details);
};
*/

template <class Host, typename Attrs> 
unsigned StateWatchers<Host, Attrs>::RegisterWatcher(const AttrSet_t& _attrs, const ObjChanged_t& _cbk, const bool& _not_in)
{
	typename StateWatchers<Host, Attrs>::WatcherDetailsPtr_t _details = new typename StateWatchers<Host, Attrs>::WatcherDetails(_attrs, _cbk, _not_in);
	return watchers.Register(_details);
}


template <class Host, typename Attrs> 
void StateWatchers<Host, Attrs>::UnRegisterWatcher(const unsigned& _id)
{
  if(do_watch_in_progress) watchers_to_remove.push_back(_id);	
  else watchers.UnRegister(_id);
};	

template <class Host, typename Attrs> 
void StateWatchers<Host, Attrs>::operator()(Host& _host, const Attrs& _attr, const std::string& _value)
{
  DoWatch func(_host, _attr, _value);
  do_watch_in_progress = true;
  watchers.for_each(typename Registrator<WatcherDetailsPtr_t>::ObjHandler_t(func));
  do_watch_in_progress = false;
  if(!watchers_to_remove.empty()) {
	typedef std::vector<unsigned>::const_iterator CI;
	for(CI p=watchers_to_remove.begin();p!=watchers_to_remove.end();++p) watchers.UnRegister(*p);
	watchers_to_remove.clear();
  };	  
};

template <class Host, typename Attrs> 
void StateWatchers<Host, Attrs>::clear()
{
  watchers.clear();	
};


//************************************ SkypeObjRegExParser ***********************************************//
template <typename Attrs> 
const std::string SkypeObjRegExParser<Attrs>::ParseResponse(const std::string& _resp)
{
  bool r = boost::regex_match(_resp.c_str(), match, re.resp);
  Assert(r);
  Assert(match.size() > re.resp_value_pos);
  
  return match[re.resp_value_pos].str();
};	

template <typename Attrs> 
const typename SkypeObjRegExParser<Attrs>::AttrValuePair_t& SkypeObjRegExParser<Attrs>::ParseEvent(const boost::cmatch& _event)
{
  parse_event_result.first  = attrs_desc[_event[re.event_attr_pos].str()]; 
  parse_event_result.second = _event[re.event_value_pos].str();
  
  return parse_event_result;	
};	

template <typename Attrs> 
void SkypeObjRegExParser<Attrs>::SetRegEx(const struct SkypeObjRegEx& _regex)
{ re = _regex; };

//************************************ SkypeObject********************************************************//
template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline 
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::SkypeObject(SkypeConnection& _sc, Host& _host, 
									    const std::string& _obj_name, 
									    const Id_t& _id, const LogFunc_t& _log_func) 
		: LogAbility(_log_func), cmd_formatter(_host, _sc), 
		  id(_id), host(_host), sc(_sc), obj_name(_obj_name), ignore_watchers(false) 
{ eh = ObjErrHandler_t(this, &SkypeObject::DummyErrHandler); };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::SkypeObject(SkypeConnection& _sc, Host& _host, 
									    const std::string& _obj_name, 
									    const Id_t& _id, const ObjErrHandler_t& _eh, 
									    const LogFunc_t& _log_func)
		: LogAbility(_log_func), cmd_formatter(_host, _sc),
		  id(_id), host(_host), sc(_sc), eh(_eh), obj_name(_obj_name),  ignore_watchers(false) 
{};

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
const std::string& SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::operator[](const Attrs& _attr) const
{
  const_iterator p = attrs.find(_attr);
  if (p == attrs.end())
    {
       LoggAssert<Id_t>(_attr,GetId());    
       Assert(p != attrs.end());
    }
  return p->second;
};

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline
typename SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::const_iterator 
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::find(const Attrs& _attr) const
{ return attrs.find(_attr); };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline
typename SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::const_iterator 
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::begin() const 
{ return attrs.begin(); };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline
typename SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::const_iterator 
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::end() const 
{ return attrs.end(); };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline	   
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::AskAttr(const Attrs& _attr)
{ AskAttr(_attr, eh); };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::AskAttr(const Attrs& _attr, const ObjErrHandler_t& _eh)
{
  cmd_formatter.FormatGetCmd(_attr);
  sc << Flush(RespHandlerImpl(_attr, host), ErrHandlerImpl(_eh, _attr, host));
};	

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
template <typename T> 
inline void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::AskForSetAttr(const Attrs& _attr, const T& _value)
{ AskForSetAttr(_attr, _value, eh); };	

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
template <typename T> 
inline void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::AskForSetAttr(const Attrs& _attr, const T& _value,
												const ObjErrHandler_t& _eh)
{
  cmd_formatter.FormatSetCmd(_attr, _value);
  sc << Flush(RespHandlerImpl(_attr, host), ErrHandlerImpl(_eh, _attr, host));
};

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline	   
const Id_t& SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::GetId() const
{ return id; };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline	   
const std::string& SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::GetName() const
{ return obj_name; };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline	   
typename SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::Watchers_t& 
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::GetWatchers()
{ return watchers; };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline	
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::IgnoreWatchers(bool _ignore)
{
  ignore_watchers = _ignore;	
};

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
SkypeConnection& SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::GetSkype()
{ return sc; };	

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
inline	   
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::ProcessEvent(const boost::cmatch& _event)
{ SetAttrValue(parser.ParseEvent(_event)); };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   >
inline	   
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::DummyErrHandler(Host& _me, const Attrs& _attr,
										     const int& _err_code, 
						  				     const std::string& _err_txt, 
										     const std::string& _err_msg)
{ log(LlError, "%s [%s/%s]: %s", __FUNCTION__, obj_name.c_str(), attrs_desc[_attr].c_str(), _err_msg.c_str()); };

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::SetAttrValue(const typename Parser_t::AttrValuePair_t& _attr_value_pair)
{
  typename std::map<Attrs, std::string>::iterator p = attrs.find(_attr_value_pair.first);

	  
  bool not_found = (p == attrs.end());
  bool r =(not_found||(p->second != _attr_value_pair.second));

  if(!not_found) p->second = _attr_value_pair.second;
  else attrs.insert(_attr_value_pair);
	 
  if(r&&(!ignore_watchers)) {
	  watchers(host, _attr_value_pair.first, _attr_value_pair.second);
  }
};

//************************************ SkypeObject::HandlerImpl*******************************************//

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   >
inline	   
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::HandlerImpl::HandlerImpl(const Attrs& _attr, Host& _host)
		: attr(_attr), host(_host)
{};

//************************************ SkypeObject::ErrHandlerImpl****************************************//
template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   >
inline	   
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::ErrHandlerImpl::ErrHandlerImpl(const ObjErrHandler_t& _eh,
												const Attrs& _attr,
												Host& _host)
		: HandlerImpl(_attr, _host), eh(_eh)
{};
		
template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   >
inline	   
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::ErrHandlerImpl::operator()(const int& _err_code, 
												const std::string& _err_txt,
												const std::string& _err_msg) 
{ eh(HandlerImpl::host, HandlerImpl::attr, _err_code, _err_txt, _err_msg);  };

//************************************ SkypeObject::RespHandlerImpl***************************************//
template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   >
inline	   
SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::RespHandlerImpl::RespHandlerImpl(const Attrs& _attr,
												 Host& _host)
		: HandlerImpl(_attr, _host) 
{};

template   < 	class Host,
		typename Attrs, 
		template <typename, class> class SkypeObjCmdFormatter,
		template <typename> class SkypeObjParser,
		typename Id_t
	   > 
void SkypeObject<Host, Attrs, SkypeObjCmdFormatter, SkypeObjParser, Id_t>::RespHandlerImpl::operator()(const std::string& _resp)
{ 
  std::string value = HandlerImpl::host.parser.ParseResponse(_resp);
  typename Parser_t::AttrValuePair_t result(HandlerImpl::attr, value);
  HandlerImpl::host.SetAttrValue(result);
};  

//************************************ SkypeObjManager::Listener******************************************//

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
inline 
SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::Listener::Listener(const AttrSet_t& _attrs, 
										  const ObjCreated_t& _cbk, 
										  const unsigned& _id) 
		: attrs(_attrs), callback(_cbk), id(_id)
{};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
inline 
SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::ListenerIdEqual::ListenerIdEqual(const unsigned& _id)
		: id(_id)
{};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
inline 
bool SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::ListenerIdEqual::operator()(const ListenerPtr_t& _p) const
{ return _p->id == id; };

//************************************ SkypeObjManager::ReleasedObjData***********************************//

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
inline 
SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::ReleasedObjData::ReleasedObjData(const ObjPtr_t& _p)
							      : object(_p), tm()
{};		

//************************************ SkypeObjManager ***************************************************//

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::SkypeObjManager(SkypeConnection& _sc, 
								  Host& _host, const LogFunc_t& _log_func)
					 	: LogAbility(_log_func),
						  SearchPolicy<Parent_t, SearchTypes>(_host, _sc, _log_func),
						  sc(_sc), host(_host), object_eh_ptr(0), 
						  obj_watch_id(0), check_listeners_run(false)  
{}; 

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::SkypeObjManager(SkypeConnection& _sc, 
							  	  Host& _host,
							  	  const typename ObjHost::ObjErrHandler_t& _eh,
							  	  const LogFunc_t& _log_func) 
					 : LogAbility(_log_func),
					   SearchPolicy<Parent_t, SearchTypes>(_host, _sc, _log_func),
					   sc(_sc), host(_host), obj_watch_id(0),
					   check_listeners_run(false)
{
  object_eh_ptr = new typename ObjHost::ObjErrHandler_t;
  *object_eh_ptr = _eh;
};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
unsigned SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::RegisterListener(const AttrSet_t& _attrs, const ObjCreated_t& _cbk)
{
  ListenerPtr_t l = new Listener(_attrs, _cbk, numerator.IncrNum());
  listeners.push_back(l);

  // Updating required_attrs set
  UpdateRequiredAttrs(_attrs);

  return l->id;
};	

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::UnRegisterListener(unsigned _id)
{
  if(check_listeners_run) {
	listeners_to_remove.push_back(_id);
	return;
  };
  
  RemoveListeners(listeners, _id);	
  
  typedef typename std::map<Id_t, ListenersVec_t>::iterator IT;
  for(IT p = waiting_listeners.begin();p != waiting_listeners.end();++p) {
	RemoveListeners(p->second, _id);
  };

  /// Remove empty vectors from the map
  for(IT ne = waiting_listeners.begin();ne!=waiting_listeners.end();) {
	if(ne->second.empty()) {
		waiting_listeners.erase(ne);
		ne = waiting_listeners.begin();
	} else ++ne;	
  };

  // Rebuilding required_attrs set
  required_attrs.clear();
  typedef typename ListenersVec_t::const_iterator CI;
  for(CI p = listeners.begin();p != listeners.end();++p) UpdateRequiredAttrs((*p)->attrs);
};	

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
typename SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::ObjPtr_t 
SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::find(const Id_t& _id, bool _want_to_own)
{
  typedef typename std::map<Id_t, ObjPtr_t>::iterator IT;
  IT p = objects.find(_id);
  if(p == objects.end()) return 0;

  if(_want_to_own) obj_refs[_id]++; 
  
  return p->second;	  
};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::ReleaseObj(const ObjHost& _obj)
{
  const Id_t id = _obj.GetId(); 	
  typedef typename std::map<Id_t, int>::iterator IT;
  IT p = obj_refs.find(id);
  if(p == obj_refs.end()) return;

  if(--p->second <= 0) {	// There are no application's objects which refers to this Skype object
	typename std::map<Id_t, ReleasedObjData>::const_iterator pp = released_ids.find(id);
	Assert(pp == released_ids.end());
	typename std::map<Id_t, ObjPtr_t>::iterator po = objects.find(id);
	Assert(po != objects.end());
	released_ids[id] = ReleasedObjData(po->second);
	po->second->IgnoreWatchers();
	objects.erase(po);
	
	typename std::map<Id_t, ListenersVec_t>::iterator pw = waiting_listeners.find(id);
	if(pw != waiting_listeners.end())  waiting_listeners.erase(pw);
	obj_refs.erase(p);
  };
};	

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::UnReleaseObj(const Id_t& _id)
{
   typedef typename std::map<Id_t, ReleasedObjData>::iterator IT;
   IT i = released_ids.find(_id);
   if (i!=released_ids.end()) released_ids.erase(i);
};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::Squeeze()
{
  std::vector<Id_t> ids;
  typedef typename std::map<Id_t, ReleasedObjData>::const_iterator CI;
  for(CI p = released_ids.begin();p != released_ids.end();++p)  
	  					if(p->second.object->IsAlive() == ObjHost::finished) ids.push_back(p->first);
  typedef typename std::vector<Id_t>::const_iterator CII;
  for(CII p = ids.begin();p != ids.end();++p) released_ids.erase(*p);

  unsigned qty = released_ids.size();
  if(qty > 1000) {
	log(LlWarning, "SkypeObjManager(%s) : %d objects in released_ids, memory leak is possible", typeid(host).name(), qty);  
  };	  
};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::AddEvent(const std::string& _regex)
{
  sc.RegisterCallback(_regex, RegExNotify_t(this, &Parent_t::EventHandler));	
};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::EventHandler(const boost::cmatch& _match)
{
  Id_t id = host.ParseEvent(_match);
 
  typename std::map<Id_t, ReleasedObjData>::iterator r_p = released_ids.find(id); 
  if(r_p != released_ids.end()) {
  	std::ostringstream oss;
  	oss << id;
	log(LlDbg, "%s: ignoring event '%s' for object %s because it is released", __FUNCTION__, _match[0].str().c_str(), oss.str().c_str());
	r_p->second.object->ProcessEvent(_match);
	return;
  };
  
  typename std::map<Id_t, ObjPtr_t>::iterator p = objects.find(id);
  
  ObjHost& obj = (p == objects.end()) ? Create(id) : *(p->second); 
  obj.ProcessEvent(_match);
 
  CheckListeners(obj);
};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
inline
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::RemoveListeners(ListenersVec_t& _vec, const unsigned& _id)
{
  typedef typename ListenersVec_t::iterator IT;
  IT ne = std::remove_if(_vec.begin(), _vec.end(), ListenerIdEqual(_id)); 	
  _vec.erase(ne, _vec.end());
};	

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::UpdateRequiredAttrs(const AttrSet_t& _attrs)
{
  typedef typename AttrSet_t::const_iterator CI;
  for(CI p = _attrs.begin();p != _attrs.end();++p) {
	Assert(attrs_desc.GetDetails(*p).can_get);
	if(attrs_desc.GetDetails(*p).reported_at_obj_creation) continue;
	if(required_attrs.find(*p) == required_attrs.end()) required_attrs.insert(*p);
  };
};	

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
inline
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::ObjChanged(ObjHost& _obj, const ObjAttrs& _attr, const std::string& _value)
{ CheckListeners(_obj); };

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
void SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::CheckListeners(ObjHost& _obj)
{
  Id_t id = _obj.GetId();	
  typename std::map<Id_t, ListenersVec_t>::iterator pp = waiting_listeners.find(id);
  check_listeners_run = true;
  
  if(pp != waiting_listeners.end()) {
	ListenersVec_t& ls = pp->second;  
	typename ListenersVec_t::size_type sz = ls.size();
	for(typename ListenersVec_t::size_type i = 0;i<sz;) {
		typedef typename AttrSet_t::const_iterator CI;
		bool all_attrs_set = true; 
		for(CI p = ls[i]->attrs.begin();p != ls[i]->attrs.end();++p) {
			if(_obj.find(*p) == _obj.end()) {
				all_attrs_set = false;
				break;
			};
		};
		if(all_attrs_set) {
			bool r = ls[i]->callback(_obj);
			if(r) obj_refs[id]++;
			ls.erase(ls.begin() + i); sz--;
		} else i++;	
	};
	if(ls.empty()) {
		waiting_listeners.erase(pp);
		_obj.GetWatchers().UnRegisterWatcher(obj_watch_id);
		if(obj_refs[id] <= 0) ReleaseObj(_obj);
	};	
  };

  check_listeners_run = false;
  if(!listeners_to_remove.empty()) {
	typedef std::vector<unsigned>::const_iterator CI;
        for(CI p=listeners_to_remove.begin();p!=listeners_to_remove.end();++p) UnRegisterListener(*p);
	listeners_to_remove.clear();
  };
};

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
unsigned SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::Size() const
{
  return objects.size();
}

template <class ObjHost, typename ObjAttrs, class Host, typename SearchTypes, template <class, class> class SearchPolicy> 
ObjHost& SkypeObjManager<ObjHost, ObjAttrs, Host, SearchTypes, SearchPolicy>::Create(Id_t _id)
{
  Assert(find(_id) == 0);
  
  ObjPtr_t obj;
  
  if(object_eh_ptr) obj = new ObjHost(sc, _id, *object_eh_ptr, GetLog());
  else obj = new ObjHost(sc, _id, GetLog());
  objects[_id] = obj;
  Assert(obj_refs.find(_id) == obj_refs.end());
  obj_refs[_id] = 0;
  waiting_listeners[_id] = listeners;

  typedef typename AttrSet_t::const_iterator CI;
  // Sending GET-commands to Skype
  for(CI p = required_attrs.begin();p != required_attrs.end();++p) obj->AskAttr(*p);
  
  typename ObjWatchers_t::ObjChanged_t f(this, &Parent_t::ObjChanged);	
  obj_watch_id = obj->GetWatchers().RegisterWatcher(AttrSet_t(), f, true);

  return *obj;
}

}; // namespace Ice
#endif // SKYPE_OBJ_H
