/*!
  \file unix/Loggers.h

  \brief Unix-specific logger (syslog) 

  $Id: Loggers.h,v 1.8 2006/05/21 13:46:25 vit Exp $

  \author (c) Ice Brains Software, llc.  http://www.icebrains-soft.com
  \date   25.07.2005 17:05:40 MSD
*/

#ifndef UNIX_SK_LOGGERS_H
#define UNIX_SK_LOGGERS_H

#include "../Cfg.h"

#ifdef SKYPE_NIX

#include "../Loggers.h"

extern "C" {
#include <syslog.h>
};

/*!
  \class SysLogger
  \brief Logging device - syslog.

  This class can be used for logging via syslog. See also documentation on #BaseLogger for additional useful remarks.
*/  
class SysLogger : public BaseLogger<SysLogger> {
public:
	/*!
  	  \brief Opens a connection to system logger 
  
  	  \param _ident		See openlog manual
  	  \param _option	See openlog manual
  	  \param _facility	See openlog manual
 	  \param  _min_lvl	Minimal log level.  Messages with log level less than _min_lvl will not be logged
	*/
	SysLogger(const std::string& _ident, int _option, int _facility, 
					const LogLevel_t _min_lvl = DefaultMinLogLevel<>::GetLvl());
	/*!
  	  \brief  Real logging function.  See BaseLogger documentation for additional information.

	  \param _lvl  Log-level. Messages with log level less than minimum log level will not be logged
	  \param _fmt  Format of log string.  See (v)printf manual for details.  
	  \param _args Arguments related to format.  See vprintf manual for additional information.
	*/
	void Do(const LogLevel_t _lvl, const char* _fmt, va_list _args);
	/// Disconnects from the syslog
	~SysLogger();
private:
	/// Ident string for syslog
	const std::string ident;
};

/**********************************************************************************************************/
/************************************ Implementation ******************************************************/
/**********************************************************************************************************/

inline void SysLogger::Do(const LogLevel_t _lvl, const char* _fmt, va_list _args)
{
  static int priorities[] = { LOG_DEBUG, LOG_INFO, LOG_WARNING, LOG_ERR,  LOG_CRIT };	
  Assert((_lvl >= LlDbg) && (_lvl <= LlCritical));
  vsyslog(priorities[_lvl], _fmt, _args);	
};

inline SysLogger::SysLogger(const std::string& _ident, int _option, int _facility, const LogLevel_t _min_lvl)
	 : BaseLogger<SysLogger>(*this, _min_lvl), ident(_ident)
{
  openlog(ident.c_str(), _option, _facility);	
};

inline SysLogger::~SysLogger()
{
  closelog();	
};

#endif // SKYPE_NIX
#endif // UNIX_SK_LOGGERS_H

